from __future__ import annotations
from typing import TYPE_CHECKING, Dict
if TYPE_CHECKING:
    from hyades.device.device import Device

from hyades.connection.base_wrapper import base_wrapper
from hyades.connection.fabric import ConnectionFabric
from genie.conf.base import Device as genie_device

@ConnectionFabric.register()
class scrapli_wrapper(base_wrapper):
    registry_name: str = 'unicon'

    def configure(self, device: Device) -> Dict[str,str]:
        self.device = device
        self.manager = self
        if device.mode == "async":
            raise UnsuportedMode("Async isn't supported for Unicon")          
        
        device_data = {
           "os": device.platform,
           "credentials": {
               'default' : {
                   'username': device.username,
                   'password': device.password
               },
           },
           'connections': {
               'a': {
                   'protocol': device.transport,
                   "ip": device.hostname,   

               },
           },
        }

        if hasattr(device, 'enable_password'):
            if device.enable_password is not None:
                device_data['credentials']['enable'] = device.enable_password

        if hasattr(device, 'port'):
            device_data['connections']['a']['port'] = device.port

        if hasattr(device, 'conn_class'):
            device_data['connections']['a']['class'] = device.conn_class
        
        self.conn = genie_device(device.name, **device_data)

        mappings = {}
        methods = ['connect', 'disconnect', 'execute', 'configure']
        for method in methods:
            mappings[method] = getattr(self, f"{device.mode}_{method}")
        
        return mappings
        
    @property
    def connected(self):
        return self.conn.connected
      
    def sync_connect(self):
        try:
            log_mode = self.device.log_mode
        except AttributeError:
            log_mode = False
        try:
            learn_mode = self.device.learn_hostname
        except AttributeError:
            learn_mode = True

        self.conn.connect(log_stdout=log_mode, learn_hostname=learn_mode)
    
    def sync_disconnect(self):
        if self.connected:
            self.conn.disconnect()    

    def sync_execute(self, cmd):
        out = self.conn.execute(cmd)
        return out
         
    def sync_configure(self, cmd):
        out = self.conn.configure(cmd)
        return out

class UnsuportedMode(Exception):
    '''
        Raises when try to use async with sync module
    '''


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from hyades.device.device import Device


class ParserFabric():
    registry = {}

    @classmethod
    def register(cls):

        def inner_wrapper(wrapped_class):
            name = wrapped_class.registry_name
            if name in cls.registry:
                raise AlreadyRegistred(f"{name} already registred")
            else:
                cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def get_parser(cls, device: Device):
        parser = device.parser
        if parser not in cls.registry:
            raise Exception(f"{parser} not registred")

        exec_class = cls.registry[parser]
        parser_manager = exec_class(device)
        return parser_manager

class AlreadyRegistred(Exception):
    '''
        Raises when tries to register an already registred parser
    '''

        


from __future__ import annotations
from typing import TYPE_CHECKING, Dict
if TYPE_CHECKING:
    from hyades.device.device import Device

from hyades.connection.base_wrapper import base_wrapper
from hyades.connection.fabric import ConnectionFabric
from scrapli import Scrapli, AsyncScrapli

@ConnectionFabric.register()
class scrapli_wrapper(base_wrapper):
    registry_name: str = 'scrapli'

    def configure(self, device: Device) -> Dict[str,str]:
        self.device = device
        self.manager = self

        device_data = {
           "host": device.hostname,
           "auth_username": device.username,
           "auth_password": device.password,
           "auth_strict_key": False,
           "platform": device.platform,
        }
        if hasattr(device, 'enable_password'):
            device_data['auth_secondary'] = device.enable_password
        
        scrapli_args = [
            'auth_private_key',
            'auth_private_key_passphrase',
            'auth_strict_key',
            'auth_bypass',
            'timeout_socket',
            'transport',
            'timeout_transport',
            'timeout_ops',
            'comms_prompt_pattern',
            'comms_return_char',
            'ssh_config_file',
            'ssh_known_hosts_file',
            'on_init',
            'on_open',
            'on_close',
            'transport_options',
            'channel_lock',
            'channel_log',
            'channel_log_mode',
            'logging_uid',
            'privilege_levels',
            'default_desired_privilege_level',
            'failed_when_contains',
            'textfsm_platform',
            'genie_platform'
        ]
        for arg in scrapli_args:
            if hasattr(device, arg):
                device_data[arg] = getattr(device, arg)
        
        if device.mode == "async":
            mode = "async"
            self.conn = AsyncScrapli(**device_data)

        else:
            mode = "sync"
            self.conn = Scrapli(**device_data)

        mappings = {}
        methods = ['connect', 'disconnect', 'execute', 'configure']
        for method in methods:
            mappings[method] = getattr(self, f"{mode}_{method}")
        
        return mappings
        
    @property
    def connected(self):
        return self.conn.isalive()
      
    async def async_connect(self):
        return await self.conn.open()

    async def async_disconnect(self):
        if self.connected:
            await self.conn.close()    

    async def async_execute(self, cmd):
        out = await self.conn.send_command(cmd)
        return out.result
         
    async def async_configure(self, cmd):
        out = await self.conn.send_configs(cmd)
        return out.result

    def sync_connect(self):
        self.conn.open()
    
    def sync_disconnect(self):
        if self.connected:
            self.conn.close()    

    def sync_execute(self, cmd):
        out = self.conn.send_command(cmd)
        return out.result
         
    def sync_configure(self, cmd):
        out = self.conn.send_configs(cmd)
        return out.result



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List
import asyncio

from hyades.parsers.fabric import ParserFabric
from hyades.connection.fabric import ConnectionFabric



class Device():
    name: str
    hostname: str
    platform: str
    username: str
    password: str
    transport: str = None
    groups: List[str] = []
    mode: str = 'async'
    parser: str = 'Default'
    connection_manager: str = 'scrapli'
    port: int = None
    enable_password: str = None
    
    def __init__(self, **kwargs) -> None:
        for attr in kwargs:
            setattr(self, attr, kwargs[attr])
        for attr in Device.__annotations__:
            if attr not in kwargs:
                try:
                    getattr(self, attr)
                except AttributeError:
                    raise MissingAttribute(f"Missing required attribute: {attr}")
                    
        self.parser_manager = ParserFabric().get_parser(self).manager
        self.connection_manager = ConnectionFabric().get_CM(self).manager

    def destroy(self):
        if self.mode == 'async':
            loop = asyncio.get_event_loop()
            loop.run_until_complete(self.disconnect())
        else:
            self.disconnect()

        del self.parser_manager
        del self.connection_manager
        

class MissingAttribute(Exception):
    ''' 
        Raises when Device __annotations__ attributes are missing when instantiating
        the object( required attributes )
    '''



from __future__ import annotations
from typing import TYPE_CHECKING, Callable, Dict

from hyades.parsers.base_parser import base_parser
if TYPE_CHECKING:
    from hyades.device.device import Device

from genie.conf.base import Device as genie_device
from hyades.parsers.fabric import ParserFabric
import asyncio

@ParserFabric.register()
class default_wrapper(base_parser):
    registry_name: str = 'Default'

    def configure(self, device: Device) -> Dict[str, Callable]:
        self.device = device
        self.manager = self
        if device.mode == "async":
            mode = "async"
        else:
            mode = "sync"
        
        platform_translation = {
            "hp_comware": 'comware',
            'cisco_iosxe': 'iosxe',
            'cisco_iosxr': 'iosxr',
            'cisco_nxos': 'nxos',
            'juniper_junos': 'junos',
        }
        if device.platform in platform_translation:
            platform = platform_translation[device.platform]
        else:
            platform = device.platform
        
        self.genie_device = genie_device(device.name,
                                         custom={"abstraction": {"order": ["os"]}},
                                         os=platform)
        mappings = {}

        mappings['parse'] =  getattr(self, f"{mode}_parse")
        return mappings
        
    def sync_parse(self, command: str) -> Dict:
        output = {}
        cmd_output = self.device.execute(command)
        output = self.genie_device.parse(command, output=cmd_output)
        return output
    
    async def async_parse(self, command: str) -> Dict:
        output = {}
        cmd_output = await self.device.execute(command)
        output = self.genie_device.parse(command, output=cmd_output)
        return output

class Unconnected(Exception):
    '''
        Raises when try to execute on unconnected Devices
    '''


from __future__ import annotations
from typing import TYPE_CHECKING, Callable, Dict, List
if TYPE_CHECKING:
    from hyades.device.device import Device


class base_parser():

    required_interfaces: List[str] = ['parse']

    def __init__(self, device: Device):
        
        mappings = self.configure(device)

        for method in base_parser.required_interfaces:
            if method not in mappings:
                raise MethodNotImplemented(f"Required method: {method} not implemented")
        
        for method in mappings:
            setattr(device, method, mappings[method])
        
    def configure(self, device: Device) -> Dict[str,Callable]:
        raise Exception("Method configure not implemented")

 

class MethodNotImplemented(Exception):
    '''
        Raises when wrapper doesn't implement
    '''


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from hyades.device.device import Device


class ConnectionFabric():

    registry = {}

    @classmethod
    def register(cls):

        def inner_wrapper(wrapped_class):
            name = wrapped_class.registry_name
            if name in cls.registry:
                raise AlreadyRegistred(f"{name} already registred")
            else:
                cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def get_CM(cls, device: Device):
        connection_manager = device.connection_manager
        if connection_manager not in cls.registry:
            raise Exception(f"{connection_manager} not registred")

        exec_class = cls.registry[connection_manager]
        conn_manager = exec_class(device)
        return conn_manager

class AlreadyRegistred(Exception):
    '''
        Raises when tries to register an already registred connection_manager
    '''




from __future__ import annotations
from typing import Any, Dict, List
from hyades.device.device import Device
import yaml

class InventoryManager():
    def __init__(self, inventory_src: Dict = None) -> None:
        self.devices: Dict[str, Device] = {}
        if isinstance(inventory_src, str):
            stream = open(inventory_src, 'r')
            yml_inv = yaml.load(stream, yaml.Loader)
            stream.close()
            devices = []
            for device in yml_inv['devices']:
                dev_dict = yml_inv['devices'][device]
                dev_dict['name'] = device
                devices.append(dev_dict)
            self.load(devices)
        elif isinstance(inventory_src, dict):
            self.load(inventory_src)        

    def load(self, inventory_dict: Dict) -> None :
        for device in inventory_dict:
            dev = Device(**device)
            if dev.name not in self.devices:
                self.devices[dev.name] = dev
            else:
                raise DuplicateError(f"Duplicate device name {dev.name}")

    def filter(self, filter_mode: str = "OR", **kwargs):
        device_names = [dev for dev in self.devices]
        return DeviceIterator(device_names, self).filter(filter_mode, **kwargs)

    def add_device(self, device_data: Any):
        if isinstance(device_data, Device):
            device_name = device_data.name
            device = device_data
        elif isinstance(device_data, Dict):
            device = Device(**device_data)
            device_name = device.name
            
        if device_name in self.devices:
            raise DuplicateError("Device {device_name} already exists")
        else:
            self.devices[device_name] = device

    def del_device(self, device_name):
        device = self.devices[device_name]
        self.devices.pop(device_name, None)
        device.destroy()
        del device
        
    def __iter__(self):
        device_names = [dev for dev in self.devices]
        return DeviceIterator(device_names, self)



class DeviceIterator():
    def __init__(self, device_names: List[str], inventory: InventoryManager) -> None:
        self.devices = inventory.devices
        self.dev_list = device_names

    def filter(self, filter_mode: str = "OR", **kwargs ):
        if filter_mode not in ["AND", "OR"]:
            raise Exception("Invalid filter_mode")
        device_names = self.dev_list
        devices = [self.devices[name] for name in device_names]
        filtred_devices = []
        filtred_devices_names = []
        for arg in kwargs:
            for device in devices:
                try:
                    if arg != "group":
                        value = getattr(device, arg)
                        if value == kwargs[arg]:
                            if device.name not in filtred_devices_names:
                                filtred_devices_names.append(device.name)
                                filtred_devices.append(device)
                    else:
                        if kwargs[arg] in device.groups:
                            if device.name not in filtred_devices_names:
                                filtred_devices_names.append(device.name)
                                filtred_devices.append(device)
                except Exception as e:
                    pass
        
        temp_filtred_devices = filtred_devices_names.copy()
        if filter_mode == "AND":
            for arg in kwargs:
                for device_name in temp_filtred_devices:
                    device = self.devices[device_name]
                    try:
                        if arg != "group":
                            value = getattr(device, arg)
                            if value != kwargs[arg]:
                                filtred_devices_names.remove(device_name)
                        else:
                            if kwargs[arg] not in device.groups:
                                filtred_devices_names.remove(device_name)
                    except Exception:
                        filtred_devices_names.remove(device_name)
        
        self.dev_list = filtred_devices_names
        return self
    
    def __iter__(self):
        return self
    
    def __next__(self):
        try:
            dev_name = self.dev_list.pop()
            return self.devices[dev_name]
        except IndexError:
            raise StopIteration



class DuplicateError(Exception):
    '''
        Raises when device has duplicated name
    '''

            


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hyades',
 'hyades.connection',
 'hyades.connection.wrappers',
 'hyades.device',
 'hyades.inventory',
 'hyades.parsers',
 'hyades.parsers.wrappers']

package_data = \
{'': ['*']}

install_requires = \
['asyncssh>=2.7.2,<3.0.0',
 'genie>=21.7,<22.0',
 'netmiko>=3.4.0,<4.0.0',
 'pyats>=21.7,<22.0',
 'scrapli-community>=2021.7.30,<2022.0.0',
 'scrapli[ssh2]>=2021.7.30,<2022.0.0']

setup_kwargs = {
    'name': 'hyades',
    'version': '0.1.0',
    'description': 'Network Automation Inventory',
    'long_description': None,
    'author': 'Renato Almeida de Oliveira',
    'author_email': 'renato.almeida.oliveira@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
from typing import TYPE_CHECKING, Dict
if TYPE_CHECKING:
    from hyades.device.device import Device

from hyades.connection.base_wrapper import base_wrapper
from hyades.connection.fabric import ConnectionFabric
from netmiko import ConnectHandler

@ConnectionFabric.register()
class scrapli_wrapper(base_wrapper):
    registry_name: str = 'netmiko'

    def configure(self, device: Device) -> Dict[str,str]:
        self.device = device
        self.manager = self
        if device.mode == "async":
            raise UnsuportedMode("Async isn't supported for Unicon")     

        device_data = {
           "host": device.hostname,
           "username": device.username,
           "password": device.password,
           "device_type": device.platform,
           "auto_connect": False,
        }
        if hasattr(device, 'enable_password'):
            device_data['secret'] = device.enable_password
        
        scrapli_args = [
            'port',
            'verbose',
            'global_delay_factor',
            'global_cmd_verify',
            'use_keys',
            'key_file',
            'pkey',
            'passphrase',
            'allow_agent',
            'ssh_strict',
            'system_host_keys',
            'alt_host_keys',
            'alt_key_file',
            'ssh_config_file',
            'conn_timeout',
            'auth_timeout',
            'banner_timeout',
            'blocking_timeout',
            'timeout',
            'session_timeout',
            'keepalive',
            'default_enter',
            'response_return',
            'serial_settings',
            'fast_cli',
            'session_log',
            'session_log_record_writes',
            'session_log_file_mode',
            'allow_auto_change',
            'encoding',            
        ]
        for arg in scrapli_args:
            if hasattr(device, arg):
                device_data[arg] = getattr(device, arg)
        
        mode = "sync"
        self.conn = ConnectHandler(**device_data)

        mappings = {}
        methods = ['connect',
                   'disconnect',
                   'execute',
                   'configure',
                   'enable',
                   'config_mode',
                   ]
        for method in methods:
            mappings[method] = getattr(self, f"{mode}_{method}")
        
        return mappings
        
    @property
    def connected(self):
        return self.conn.is_alive()
      
    def sync_connect(self):
        self.conn.establish_connection()
    
    def sync_disconnect(self):
        if self.connected:
            self.conn.disconnect()    

    def sync_enable(self):
        self.conn.enable()

    def sync_config_mode(self):
        self.conn.config_mode()

    def sync_execute(self, cmd):        
        out = self.conn.send_command(cmd)
        return out
         
    def sync_configure(self, cmd):
        out = self.conn.send_config_set(cmd)
        return out



class UnsuportedMode(Exception):
    '''
        Raises when try to use async with sync module
    '''